# MIT License
#
# Copyright 2020 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#' Map over a very large table, reading in a chunk at a time
#'
#' Produce a result output per chunk of the file or write out
#'
#' @param infile large table to which function will be applied
#' @param FUNC function to apply over each chunk of the table
#' @param chunkSize number of lines to chunk by
#' @param header flag to specify whether or not the file has a header
#' @param ARGS additional arguments to FUNC
#' @param verbose flag to produce verbose output
#' @param outfile if not NULL, result will be written directly to this location
#' @return list(output of function FUNC called on each chunk)
#' @export
chunkedMap<-function (infile, FUNC=NULL, chunkSize=2000, header=TRUE, ARGS=list(), verbose=TRUE, outfile=NULL) {
  if (tail(strsplit(infile, "."), n=1) == "gz") stop("`infile` must not be gzipped")
  #this gets the number of lines in the file quickly.
  #this includes the header.
  lines=R.utils::countLines(infile)[1]
  if (lines<(chunkSize)) chunkSize=lines

  #loop over the lines by bins.
  index=0;
  result=list()

  if (header) cols=names(fread(infile, nrow=0, header=T))

  while (((index*chunkSize)+1)<lines) {

    skipNum=(index*chunkSize)+as.numeric(header)

    if (verbose) cat (paste("Index [", index, "] skipLines [", skipNum, "]\n", sep=""))

    #if this isn't the first iteration, add +1 for the header.
    d=fread(infile, nrow=chunkSize, skip=skipNum+as.numeric(header))
    names(d) <- cols

    if (verbose) cat (paste ("Memory Used [", paste (round (pryr::mem_used()/1e6,1), "MB")  , "]"))
    index=index+1

    #DO WORK HERE
    res=do.call("FUNC", c(list(d), ARGS))
    if (is.null(outfile))
      result[[index]]=res
    else if (index > 1)
      fwrite(res, outfile, append = T)
    else
      fwrite(res, outfile, col.names = header)
  }
  if (is.null(outfile))  
    return (result)
  else
    invisible()
}

#' Quietly apply a function over a very large table, reading in one chunk at a time
#'
#' @param infile large table to which function will be applied
#' @param FUNC function to apply over each chunk of the table
#' @param chunkSize number of lines to chunk by
#' @param header flag to specify whether or not the file has a header
#' @param ARGS additional arguments to FUNC
#' @param verbose flag to produce verbose output
#' @export
chunkedSilentApply<-function (infile, FUNC=NULL, chunkSize=2000, header=TRUE, ARGS=list(), verbose=TRUE) {
  #this gets the number of lines in the file quickly.
  #this includes the header.
  lines=R.utils::countLines(infile)[1]
  if (lines<(chunkSize)) chunkSize=lines

  #loop over the lines by bins.
  index=0;
  result=list()

  if (header) cols=names(fread(infile, nrow=0, header=T))

  while (((index*chunkSize)+1)<lines) {

    skipNum=(index*chunkSize)+as.numeric(header)

    if (verbose) cat (paste("Index [", index, "] skipLines [", skipNum, "]\n", sep=""))

    #if this isn't the first iteration, add +1 for the header.
    d=fread(infile, nrow=chunkSize, skip=skipNum+as.numeric(header))
    names(d) <- cols

    if (verbose) cat (paste ("Memory Used [", paste (round (pryr::mem_used()/1e6,1), "MB")  , "]"))
    index=index+1

    FUNC(d)
  }
  invisible()
}

#' Filter a very large table by a predicate function
#'
#' Produce a subset of the larger table
#'
#' @param infile large table to be filtered
#' @param FUNC predicate function by which to filter the table
#' @param chunkSize number of lines to chunk by
#' @param header flag to specify whether or not the file has a header
#' @param ARGS additional arguments to FUNC
#' @param verbose flag to produce verbose output
#' @param outfile if not NULL, result will be written directly to this location
#' @return data.table object of filtered file
#' @export
chunkedFilter<-function (infile, FUNC=NULL, chunkSize=2000, header=TRUE, ARGS=list(), verbose=TRUE, outfile=NULL) {
  #this gets the number of lines in the file quickly.
  #this includes the header.
  lines=R.utils::countLines(infile)[1]
  if (lines<(chunkSize)) chunkSize=lines

  #loop over the lines by bins.
  index=0;
  result=list()

  if (header) cols=names(fread(infile, nrow=0, header=T))
  while (((index*chunkSize)+1)<lines) {
    skipNum=(index*chunkSize)+as.numeric(header)

    if (verbose)
      cat (paste("Index [", index, "] skipLines [", skipNum, "]\n", sep=""))

    #if this isn't the first iteration, add +1 for the header.
    d=fread(infile, nrow=chunkSize, skip=skipNum+as.numeric(header))
    names(d) <- cols

    if (verbose) cat (paste ("Memory Used [", paste (round (pryr::mem_used()/1e6,1), "MB")  , "]"))

    index=index+1

    #DO WORK HERE
    idx = do.call("FUNC", c(list(d), ARGS))
    filteredChunk = d[idx,]
    isEmpty=nrow(filteredChunk) == 0  
    if (!isEmpty) {
      if (is.null(outfile)) result[[as.character(index)]] = filteredChunk
      else if (index > 1) fwrite(filteredChunk, outfile, append = TRUE)
      else fwrite(filteredChunk, outfile, col.names = header)
    } 
  }
  if (is.null(outfile)) 
    return (rbindlist(result))
  else
    invisible()
}

#' Iterate over a very large table, updating ACC along the way
#'
#' @param infile large table to which function will be applied
#' @param FUNC function to apply over each chunk of the table.
#'        must match signature (chunk : data.table, ACC : list) -> ACC' : list
#' @param chunkSize number of lines to chunk by
#' @param header flag to specify whether or not the file has a header
#' @param ACC initial values that will eventually be returned or written
#' @param verbose flag to produce verbose output
#' @param outfile if not NULL, result will be written directly to this location
#' @return value of ACC at the end of iteration
#' @export
chunkedReduce<-function (infile, FUNC=NULL, chunkSize=2000, header=TRUE, ACC=list(), verbose=TRUE, outfile=NULL) {
  if (tail(strsplit(infile, "."), n=1) == "gz") stop("`infile` must not be gzipped")
  #this gets the number of lines in the file quickly.
  #this includes the header.
  lines=R.utils::countLines(infile)[1]
  if (lines<(chunkSize)) chunkSize=lines

  #loop over the lines by bins.
  index=0;

  if (header) cols=names(fread(infile, nrow=0, header=T))

  while (((index*chunkSize)+1)<lines) {

    skipNum=(index*chunkSize)+as.numeric(header)

    if (verbose) cat (paste("Index [", index, "] skipLines [", skipNum, "]\n", sep=""))

    #if this isn't the first iteration, add +1 for the header.
    d=fread(infile, nrow=chunkSize, skip=skipNum+as.numeric(header))
    names(d) <- cols

    if (verbose) cat (paste ("Memory Used [", paste (round (pryr::mem_used()/1e6,1), "MB")  , "]"))
    index=index+1

    #DO WORK HERE
    ACC = FUNC(d, ACC)

    if (!is.null(outfile)) {
      if (index > 1)
        fwrite(res, outfile, append = T)
      else
        fwrite(res, outfile, col.names = header)
    }
  }
  if (is.null(outfile))  
    return (ACC)
  else
    invisible()
}
