# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#' Test if string ends with given suffix
#'
#' @param theString string to be tested
#' @param theExt suffix to be tested for
#' @return TRUE if theString ends with theExt
#' @export
strEndsWith<-function(theString,theExt) {
    return(substring(theString,1+nchar(theString)-nchar(theExt))==theExt)
}

#' Test if string starts with given prefix
#'
#' @param theString string to be tested
#' @param thePrefix prefix to be tested for
#' @return TRUE if theString starts with thePrefix
#' @export
strStartsWith = function(theString, thePrefix) {
    return(substr(theString, 1, nchar(thePrefix)) == thePrefix)
}

#' Remove a suffix from a string
#'
#' It is an error if the string does not end with the suffix.
#'
#' @param str the string to be truncated
#' @param suffix the string to be checked for and removed
#' @return the string with suffix removed
#' @export
removeSuffix = function(str, suffix) {
    if (!strEndsWith(str, suffix)) {
        stop(paste0("'", str, "' does not end with '", suffix, "'"))
    }
    return(substr(str, 1, nchar(str) - nchar(suffix)))
}

#' Remove a suffix from a string, if it is present.
#'
#' If the string does not end with the suffix, it is not truncated.
#'
#' @param str the string to be truncated
#' @param suffix the string to be checked for and removed
#' @return the string with suffix removed, if present, otherwise the original string
#' @export
maybeRemoveSuffix = function(str, suffix) {
    if (!strEndsWith(str, suffix)) {
        return(str)
    } else {
        return(substr(str, 1, nchar(str) - nchar(suffix)))
    }
}

#'
#' @param str String on which to operate.
#' @param oldSuffix suffix to be replaced.
#' @param newSuffix the replacement.
#' @return str with oldSuffix replaced by newSuffix
#' @export
replaceSuffix = function(str, oldSuffix, newSuffix) {
    return(paste0(removeSuffix(str, oldSuffix), newSuffix))
}

#' Split string into pieces of the given length
#' @param s String to be split
#' @param len desired length
#' @return vector of pieces.  Note that last piece may not be full length.
#' @export
splitStringByLength<-function(s, len) {
    if (nchar(s) <= len) return(s)
    starts <- seq(1, nchar(s), len)
    ends <- seq(len, nchar(s), len)
    if (length(ends) < length(starts)) {
        ends <- c(ends, nchar(s))
    }
    return(substring(s, starts, ends))
}

#' Read a data.table using data.table fread()
#'
#' If inFile ends with .gz, it is gunzipped in a way that is robust to running out of disk space.
#' Environment variable DROPSEQ_TMPDIR should be set to a directory that has plenty of room and is writable.
#'
#' @param inFile path of file to be read
#' @param comment_regexp: If defined, filter input through "egrep -v comment_regexp"
#' @param ... passed through to data.table::fread
#' @return a data.table
#' @export
fastRead<-function(inFile, comment_regexp=NA, ...) {
    if (!file.exists(inFile)) {
        stop(paste0(inFile, " does not exist"))
    }
    # fread writes process output to /dev/shm if it exists, which is OK so long as it is big enough.
    # If that fails, fall back to function that uses an explicit TMPDIR
    if (length(grep(".gz", inFile))==1) {
        unzip_command = paste("gunzip","-c",inFile,sep=" ")
        if (!is.na(comment_regexp)) {
            unzip_command = paste0(unzip_command, sprintf(" | egrep -v '%s'", comment_regexp))
        }
        # data.table 1.11.8 emits a warning if command string isn't passed via cmd=, but v 1.10.4 (current dotkit version)
        # doesn't recognized cmd=
        if (version_above("data.table", "1.11.8")) {
            return(tryCatch(data.table::fread(cmd=unzip_command,data.table=T, ...), error=function(e) fastReadBigGz(inFile, comment_regexp=comment_regexp, ...)))
        } else {
            return(tryCatch(data.table::fread(unzip_command,data.table=T, ...), error=function(e) fastReadBigGz(inFile, comment_regexp=comment_regexp, ...)))
        }

    } else {
        input = inFile
        if (!is.na(comment_regexp)) {
            input = sprintf("egrep -v '%s' %s", comment_regexp, inFile)
            a=data.table::fread(input, data.table=T, ...)
        } else {
            a=data.table::fread(input, data.table=T, ...)
        }
        return(a)
    }
}

# internal function
fastReadBigGz<-function(inFile, ...) {
    # Grid Engine overwrites TMPDIR, so support our own environment variable that won't get clobbered.
    tmpDir = Sys.getenv('DROPSEQ_TMPDIR')
    if (nchar(tmpDir) == 0) {
        tmpDir = Sys.getenv('TMPDIR')
    }
    if (nchar(tmpDir) == 0) {
        tmpDir = tempdir()
    }
    t=tempfile(tmpdir=tmpDir, fileext=".tsv")
    on.exit(unlink(t), add = TRUE)
    cmd=paste("gunzip -c", inFile, ">", t)
    retval = system(cmd,intern=F) != 0
    if (retval != 0) {
        stop(paste0(cmd, " failed with status ", retval))
    }
    return(fastRead(t, ...))
}

# Returns a data.frame with genes as rownames, no GENE column.
# Cells (columns) are in decreasing order by size.
read_dge_gz<-function(file, decreasing_order_by_size=TRUE) {
    dge = fastRead(file, comment_regexp = '^#')
    setkey(dge,GENE)
    gene_names = dge$GENE
    dge[, GENE:=NULL]
    rownames(dge) = gene_names

    if (decreasing_order_by_size) {
        # order cells by expression
        colsums.full.data=dge[,lapply(.SD,sum,na.rm=TRUE),.SDcols=colnames(dge)]
        order.data=names(colsums.full.data)[order(colsums.full.data,decreasing=T)]
        full.data.cells=dge[,order.data,with=F]
    } else {
        full.data.cells = dge
    }
    full.data.cells = as.data.frame(full.data.cells)
    rownames(full.data.cells) = gene_names
    return(full.data.cells)
}



RawDgeMultiplier = 10000

#' this function converts normalized data back to DGEs (can be useful for performing counts-based analyses)
#' rownames of input must be gene names.  If input has GENE column, it is removed.
#' @param a log-normalized DGE
#' @param returnDataFrame if TRUE, returns a data.frame with gene rownames; else returns a data.table with GENE column and gene rownames.
#' @export
ConvertNormalizedToDGE = function(a, returnDataFrame=TRUE) {
    genes=row.names(a)

    aa = setDT(a)
    if ('GENE' %in% colnames(aa)) {
        aa[, GENE:=NULL]
    }
    #check to see if a column contains no 0 values.  This is false if not ok.
    idxOK=apply(a, 2, function (x) any(which(x==0)))
    if (any(idxOK==F)) warning("Some cells had no 0 values for expression.  You need at least one gene with 0 expression to correctly recover raw counts")

    aa = (exp(aa)-1)/RawDgeMultiplier

    fixColumn<-function (x) {
        txptsum<-1/(sort(unique(x))[2])
        round(x*txptsum)
    }

    for (col in colnames(aa)) {
        set(aa, j=col, value= fixColumn (aa[[col]]))
    }

    if (returnDataFrame) {
        setDF(aa)
        rownames(aa)=genes
        return(aa)
    } else {
        return(addGeneColumnToDgeDt(aa, genes))
    }
}

# internal function
addGeneColumnToDgeDt = function(dt, genes) {
    cell_barcodes = colnames(dt)
    dt$GENE=genes
    setcolorder(dt, c('GENE', cell_barcodes))
    rownames(dt)=genes
    return(dt)
}

#' Log-normalize a DGE data frame or data.table
#'
#' @param a Raw(integer) DGE data fram or data.table. rownames of input must be gene names.  If input has GENE column,
#'       it is removed.
#' @param returnDataFrame If true, returns a data.frame with gene rownames; else returns a data.table with GENE column
#'       and gene rownames.
#' @return A data frame (or data.table) in which the expression for each cell is normalized so it sums to 1, and then
#'      logged.
#' @import data.table
#' @export
ConvertDgeToNormalized = function(a, returnDataFrame = TRUE) {
    genes=rownames(a)
    a = setDT(a)
    if ('GENE' %in% colnames(a)) {
        a[, GENE:=NULL]
    }
    a = log(RawDgeMultiplier*sweep(a,2,colSums(a),'/')+1)

    if (returnDataFrame) {
        setDF(a)
    rownames(a) = genes
    return(a)
    } else {
        setDT(a)
        return(addGeneColumnToDgeDt(a, genes))
    }
}

#' normalize a DGE data frame or data.table, so that each cell's expression sums to 1
#'
#' @param a Raw(integer) DGE data fram or data.table. rownames of input must be gene names.  If input has GENE column,
#'      it is removed.
#' @param returnDataFrame If true, returns a data.frame with gene rownames; else returns a data.table with GENE column
#'      and gene rownames.
#' @return A data frame (or data.table) in which the expression for each cell is normalized so it sums to 1.
#' @import data.table
#' @export
ConvertDgeToNormalizedNoLog = function(a, returnDataFrame = TRUE) {
    genes=rownames(a)
    a = setDT(a)
    if ('GENE' %in% colnames(a)) {
        a[, GENE:=NULL]
    }
    a = sweep(a,2,colSums(a),'/')
    #cs=colSums(a)
    #for (i in 1:length(cs)) {
    #    a[,i]=a[,i, with=FALSE]/cs[i]
    #}

    if (returnDataFrame) {
        setDF(a)
    rownames(a) = genes
    return(a)
    } else {
        setDT(a)
    return(addGeneColumnToDgeDt(a, genes))
    }
}

#' Log-transform a normalized DGE
#'
#' Handles 0 entries by multiplying by 10000 and adding 1.
#'
#' @param normalizedDge a DGE (without GENE column) as produced by ConvertDgeToNormalizedNoLog
#' @return a DGE as produced by ConvertDgeToNormalized
#' @export
ScaleAndLogDge = function(normalizedDge) {
    return(log(RawDgeMultiplier*normalizedDge + 1))
}

#' Creates a file connection with the given open mode.
#' @param file If ends with ".gz", a gzfile() is created; else a regular file() connection.
#' @export
open_conn = function(file, open="") {
    if (strEndsWith(file, ".gz")) {
        # work around bug in gzfile:
        # https://stackoverflow.com/questions/45665496/how-would-one-readlines-from-a-gzip-file-in-r
        if (nchar(open) == 0) {
            open = "rb"
        }
        if (!strEndsWith(open, "b")) {
            open=paste0(open, "b")
        }
        return(gzcon(file(file, open=open)))
    } else {
        return(file(file, open=open))
    }
}

#' Write a tab-separated file with row and column labels, including a label for the rownames.
#' Additional args are passed to write.table
#' @param x the data frame to be written.
#' @param file where to write the data
#' @param the column header for the rownames
#' @export
write_table_with_label_for_rownames = function(x, file, rowname_label, quote=FALSE, sep="\t", col.names=TRUE, ...) {
    conn = open_conn(file=file, open="w")
    cat(paste0(rowname_label, '\t'), file=conn)
    utils::write.table(x, conn, col.names=col.names, row.names=TRUE, quote=quote, sep=sep, ...)
    close(conn)
}

#' @export
read_table_with_label_for_rownames=function(file, ...) {
    ret = fastRead(file, ...)
    rowname_column = colnames(ret)[1]
    labels = ret[, get(rowname_column)]
    ret[, (rowname_column):=NULL]
    ret = as.data.frame(ret)
    rownames(ret) = labels
    return(ret)
}

#' @export
ensure_trailing_slash=function(directory) {
    if (!strEndsWith(directory, "/")) {
        directory = paste0(directory, "/")
    }
    return(directory)
}

#' Check if a directory exists, and create it if not.  Errors if cannot be created or exists but is not a directory.
#' @return the directory, always with a trailing slash.  Raises an error if cannot be created
#' @export
ensure_directory_exists = function(directory, recursive=TRUE) {
    directory = ensure_trailing_slash(directory)
    if (dir.exists(directory)) {
    return(directory);
    } else if (file.exists(directory)) {
        stop(paste0("Cannot create directory because there is an existing file with the same name: ", directory))
    } else if (!dir.create(directory, recursive=recursive)) {
        stop(paste0("Error creating directory: ", directory))
    } else {
        return(directory)
    }
}

#' Convert data frame stored in RDS file to a tab-separated file
#' @param rds.file the file to be read
#' @param tsv.file the file to be written.  If ends with .gz, file is gzipped.
#' @param rowname_label The column header for the rownames
#' @export
convert_rds_to_tsv = function(rds.file, tsv.file, rowname_label, ...) {
    df = readRDS(rds.file)
    write_table_with_label_for_rownames(df, file=tsv.file, rowname_label=rowname_label, ...)
}

#' Count the number of lines at the beginning of the (possible gzipped) file that start with '@'
#' @export
countHeaderLines = function(path) {
    conn = file(path, "r")
    on.exit(close(conn), add = TRUE)
    headerLength = 0
    while(strStartsWith(line<-(readLines(conn, n=1)[[1]]), '@') || strStartsWith(line, '#')) {
        headerLength = headerLength + 1
    }
    return(headerLength)
}

dge_header_attribute_name='Drop-seq.dge.header'

version_above <- function(pkg, than) {
    as.logical(compareVersion(as.character(packageVersion(pkg)), than))
}

#' @export
read_dge_txt = function(path) {
    headerLength = countHeaderLines(path)
    dge = fastRead(path, skip=headerLength)
    if (headerLength > 0) {
        header = readLines(path, n=headerLength)
    attr(dge, dge_header_attribute_name) = header
    }
    return(dge)
}

#' @param lines vector of DGE header lines.
#' @param file If is.null(lines), this entire file is read for lines (which will be way too much if the file contains
#'             more than just the DGE header).
#' @param headerLineIntroChar Each line of the header must start with one of these characters.
#' @return A list of lists of lists.  The top-level list is a grouping by header record type, typically DGE and LIBRARY.
#'         Each sub-list is a contains records of the same type, with each sub-sub-list being list(FIELDNAME=VALUE)
#' @export
parseDgeHeader<-function(file=NULL, lines=NULL, headerLineIntroChar=c('#', '@')) {
    if (is.null(lines)) {
        lines = readLines(file)
    }
    parseHeaderLine<-function(line) {
        if (!substr(line, 1, 1) %in% headerLineIntroChar) {
            stop(paste("Unexpected line in DGE header:", line))
        }
        line=substr(line, 2, nchar(line))
        fields=strsplit(line, "\t", fixed=TRUE)[[1]]
        lineType=fields[1]
        fields=fields[2:length(fields)]
        l = list()
        rexp <- "^([^:]+):(.*)$"
        for (field in fields) {
            l[[sub(rexp, "\\1", field)]] = sub(rexp, "\\2", field)
        }
        ret=list()
        ret[[lineType]] = l
        return(ret)
    }
    parsedLines = lapply(lines, parseHeaderLine)
    lineTypes = unique(sapply(parsedLines, function(parsedLine) names(parsedLine)))
    ret = list()
    for (lineType in lineTypes) {
        selector = sapply(parsedLines, function(parsedLine) names(parsedLine) == lineType)
        ret[[lineType]] = lapply(parsedLines[selector], function(l) l[[lineType]])
    }
    return(ret)
}

#' Convert parsed DGE header record into text form (without introductory comment character)
#' @param recordType e.g. "DGE" or "LIBRARY"
#' @param list of name=value from a sub-sub-list of parseDgeHeader return value
#' @return string representation of the record.
#' @export
unparseDgeHeaderRecord<-function(recordType, rec) {
    paste(recordType, lapply(names(rec), function(n) sprintf("%s:%s", n, rec[[n]])), sep="\t", collapse="\t")
}


#' Determine if a file is in matrix market file format.
#'
#' @param file The file to test
#'
#' @return true if the file is in matrix market format
#' @export
isMatrixMarket<-function(file) {
    conn=file(file, "r")
    line=readLines(con=conn, n=1, ok=FALSE)
    close(conn)
    return(strStartsWith(line, "%%MatrixMarket matrix coordinate"))
}

#' returns list(genes, cell_barcodes)
loadSparseDgeNames=function(file) {
    conn=file(file, "r")
    genes=c()
    cell_barcodes=c()
    while(TRUE) {
        line=readLines(con=conn, n=1, ok=FALSE)
        if (!strStartsWith(line, '%')) {
            break
        }
        if (strStartsWith(line, '%%GENES\t')) {
            these_genes=strsplit(line, "\t", fixed=TRUE)[[1]]
            genes = append(genes, these_genes[2:length(these_genes)])
        } else if (strStartsWith(line, '%%CELL_BARCODES\t')) {
            these_cells=strsplit(line, "\t", fixed=TRUE)[[1]]
            cell_barcodes = append(cell_barcodes, these_cells[2:length(these_cells)])
        }
    }
    close(conn)
    return(list(genes=genes, cell_barcodes=cell_barcodes))
}

#' @return list(numRows, numCols, numPopulatedElements)
#' @export
getSparseMatrixDimensions=function(file) {
    conn=open_conn(file, "r")
    line = "%"
    while(strStartsWith(line, '%')) {
        line=readLines(con=conn, n=1, ok=FALSE)
    }
    fields = strsplit(line, "\t", fixed=TRUE)[[1]]
    return(list(numRows=as.numeric(fields[1]), numCols=as.numeric(fields[2]), numPopulatedElements=as.numeric(fields[3])))
}

#' @return list(numRows, numCols)
#' @export
getDenseDgeDimensions=function(file) {
    numCols = ncol(fastRead(file, nrows=1)) - 1
    # start at -1 because first line is a header
    numRows = -1
    conn = open_conn(file)
    while (TRUE) {
        l = readLines(conn, ok=TRUE, n=1)
        if (length(l) == 0) {
            break
        }
        if (!strStartsWith(l, "#")) {
            numRows = numRows + 1
        }
    }
    close(conn)
    return(list(numRows=numRows, numCols=numCols))
}

#' Determine size of sparse or dense DGE
#' @return list(numRows, numCols, numPopulatedElements)
#' @export
getDgeDimensions<-function(file) {
    if (isMatrixMarket(file)) {
        return(getSparseMatrixDimensions(file))
    } else {
        ret = getDenseDgeDimensions(file)
        ret$numPopulatedElements = ret$numRows * as.numeric(ret$numCols)
        return(ret)
    }
}

#' Load a sparse DGE (either real or integer), that may have gene names and cell_barcode names in header.
#'
#' @param file path to file to be loaded, in Matrix Market form.
#' @param column_compressed If TRUE, return dgCMatrix instead of dgTMatrix
#' @return A sparse matrix in triplet form (dgTMatrix), unless column_compressed
#' @export
loadSparseDge = function(file, column_compressed=FALSE) {
    genes_and_cell_barcodes = loadSparseDgeNames(file)
    ret=Matrix::readMM(file)
    rownames(ret)=genes_and_cell_barcodes$genes
    colnames(ret)=genes_and_cell_barcodes$cell_barcodes
    if (column_compressed) {
        return(convertTripletCompressedMatrixToColumnCompressed(ret))
    } else {
        return(ret)
    }
}

# Convert a sparse matrix from triplet form to column-compressed form
convertTripletCompressedMatrixToColumnCompressed = function(triplet_matrix) {
    if (! ('dgTMatrix' %in% methods::is(triplet_matrix))) {
        stop("Argument is not instance of dgTMatrix")
    }
    return(Matrix::sparseMatrix(i=triplet_matrix@i, j=triplet_matrix@j, x=triplet_matrix@x, dim=dim(triplet_matrix), dimnames=dimnames(triplet_matrix), index1=FALSE))
}

#' Detect whether file is sparse or dense, and load as appropriate
#'
#' Reads either Drop-seq Matrix Market sparse format, or tab-separate dense format.
#' Either can be gzipped.
#'
#' @param file path of file to be loaded
#' @param decreasing_order_by_size affects dense DGE only, so leaving it FALSE gives most consistent behavior.
#' @export
loadSparseOrDenseDge<-function(file, decreasing_order_by_size=FALSE) {
    if (isMatrixMarket(file)) {
        return(loadSparseDge(file))
    } else {
        return(read_dge_gz(file, decreasing_order_by_size=decreasing_order_by_size))
    }
}

removeElementFromDotList<-function (dotList) {
    a = list(...)
    str(a)
    a$toRemove = NULL
    str(a)
}

#' @export
MatrixMarketConstants = list(
    MM_COMMENT_LINE_START = "%",
    MM_STRUCTURED_COMMENT_LINE_START = "%%",
    MM_MATRIX_TYPE_GENERAL="general",
    MM_HEADER_LIST_SEPARATOR = "\t",
    ROWS = "ROWS",
    COLS = "COLS",
    GENES = "GENES",
    CELL_BARCODES = "CELL_BARCODES",
    NUM_HEADER_ELEMENTS_PER_ROW = 1000,
    real = "real",
    integer = "integer",
    real_element_format = "%d\t%d\t%.8g",
    integer_element_format = "%d\t%d\t%d"
)
MatrixMarketConstants$MM_HEADER_START = paste0(MatrixMarketConstants$MM_STRUCTURED_COMMENT_LINE_START, "MatrixMarket matrix coordinate")

writeSparseDgeLabels<-function(conn, labelType, labels) {
    label_chunks = split(labels, ceiling(seq_along(labels)/MatrixMarketConstants$NUM_HEADER_ELEMENTS_PER_ROW))
    lapply(label_chunks, function(chunk) {
        label_intro = paste0(MatrixMarketConstants$MM_STRUCTURED_COMMENT_LINE_START, labelType)
        writeLines(paste(c(label_intro, chunk), collapse="\t"), con=conn)
    })
}

#' Write DGE in Drop-seq sparse Matrix Market format
#'
#' Writes sparse matrix with Drop-seq extension, i.e. rownames and colnames in the comments.
#' Output is gzipped if file ends with '.gz.
#'
#' @param dge triplet sparse matrix, i.e. dgTMatrix
#' @param file path to file to be written
#' @param elementType 'real' or 'integer'
#' @export
saveSparseDge = function(dge, file, elementType=MatrixMarketConstants$real) {
    conn = open_conn(file, "w")
    # MatrixMarket header
    writeLines(paste(MatrixMarketConstants$MM_HEADER_START, elementType, MatrixMarketConstants$MM_MATRIX_TYPE_GENERAL, sep=" "),
        con=conn)

    # Drop-seq augmented comments that hold rownames and colnames
    writeSparseDgeLabels(conn, MatrixMarketConstants$GENES, rownames(dge))
    writeSparseDgeLabels(conn, MatrixMarketConstants$CELL_BARCODES, colnames(dge))

    # Data header
    writeLines(paste(nrow(dge), ncol(dge), length(dge@x), sep="\t"), con=conn)
    if (elementType == MatrixMarketConstants$real) {
        format_str = MatrixMarketConstants$real_element_format
    } else {
        format_str = MatrixMarketConstants$integer_element_format
    }
    for (i in 1:length(dge@x)) {
        writeLines(sprintf(format_str, dge@i[i]+1, dge@j[i]+1, dge@x[i]), con=conn)
    }
    close(conn)
}

#' @export
cat_with_time = function(...) {
    cat(paste0(Sys.time(), ": "), ...)
cat("\n") # Force newline to be printed
}

#' Invoke list.files, and throw exception if number of files matched != 1
#' @param description of type of file being sought, for error messages
#' @param other arguments are same as for list.files
#' @export
get_single_file_matching_regexp=function(description, path=".", ...) {
    files = list.files(path=path, ...)
if (length(files) > 1) {
    stop(sprintf("More than one %s file found in %s", description, path))
} else if (nchar(files) == 0) {
    stop(sprintf("No %s files found in %s", description, path))
}
return(files)
}


#' Read an rdata file and return the nth object in it without clobbering the caller's environment.
#' @export
load_obj <- function(f, whichObject=1) {
    env <- new.env()
    nm <- load(f, env)[whichObject]
    env[[nm]]
}

#' Call system2 with stdout=TRUE, and stop() if non-zero exit status
#' Arguments as in system2()
#' @return stdout of command
#' @export
system2_capture_stdout_check_error=function(command, args=character(), ...) {
    ret = system2(command=command, args=args, stdout=TRUE, ...)
    status = attr(ret, 'status')
    if (!is.null(status) && status != 0) {
        full_errmsg = sprintf("Exit status %d from '%s %s'", status, command, paste(args, collapse=" "))
        errmsg = attr(ret, 'errmsg')
        if (is.null(errmsg)) {
            full_errmsg = paste0(full_errmsg, " : ", errmsg)
        }
        stop(full_errmsg)
    } else {
        return(ret)
    }
}
