# MIT License
#
# Copyright 2017 Broad Institute
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

#' Get a value (typically from a yaml object), or default_val if val is null
#' @export
get_or_else = function(val, default_val) {
    if (is.null(val)) {
        return(default_val)
    } else {
        return(val)
    }
}

#' Compare 2 sets of assignments for mostly-same elements
#'
#' Produce a confusion matrix of the levels in each of the inputs.
#'
#' @param a,b are both factors, but it is not known whether the levels of a and the levels of b are equivalent
#' @return list(confusion_matrix of size of intersection between corresponding levels of a and b, as a fraction of the size of b for that level,
#'               list of values in a only,
#'               list of values in b only)
#' @export
make_class_intersection_matrix=function(a, b) {
    common_names = intersect(names(a), names(b))
    a_only = setdiff(names(a), common_names)
    b_only = setdiff(names(b), common_names)
    a = a[common_names]
    b = b[common_names]
    ret = matrix(nrow=length(levels(a)), ncol=length(levels(b)), dimnames=list(levels(a), levels(b)))
    for (a_level in levels(a)) {
        this_a = names(a[a==a_level])
        for(b_level in levels(b)) {
            this_b = names(b[b==b_level])
            num_this_b = length(this_b)
            ret[a_level, b_level] = length(intersect(this_a, this_b))/num_this_b
        }
    }
    return(list(confusion_matrix=ret, a_only=a_only, b_only=b_only))
}

#' Load a dictionary of values into caller's environment
#' @param dict list in which names become variable names, and values become variable values
#' @param envir Where to load the assignments
#' @export
loadEnvironment<-function(dict, envir=parent.frame()) {
    for (n in names(dict)) {
        assign(n, dict[[n]], envir)
    }
}

#' Prepare a matrix filled with zeroes in the appropriate dimensions for metacells
create_sum_by_cluster_matrix<-function(idents, raw_dge, expectedClass) {
    if (!is(raw_dge, expectedClass)) {
        stop(sprintf("class(raw_dge) (%s) != %s "), class(raw_dge), expectedClass)
    }
    if (length(idents) != ncol(raw_dge)) {
        stop("length(idents) != ncol(raw_dge)")
    }
    m = matrix(numeric(length=nlevels(idents) * nrow(raw_dge)), ncol=nlevels(idents), nrow=nrow(raw_dge))
    rownames(m) = rownames(raw_dge)
    colnames(m) = levels(idents)
    return(m)
}

# ' Create metacells from dgCMatrix sparse matrix
sum_dgCMatrix_dge_by_cluster<-function(idents, raw_dge) {
    m = create_sum_by_cluster_matrix(idents, raw_dge, "dgCMatrix")
    x = raw_dge@x # the non-zero elements of the sparse matrix
    i = raw_dge@i # zero-based row number of corresponding value in x
    p = raw_dge@p # nth element of p is the zero-based index into x of the first non-zero element of column n
    for (in_column in 1:length(idents)) {
        out_column = as.integer(idents[in_column])
        for (value_index in (p[in_column]+1):(p[in_column+1])) {
            m[i[value_index]+1, out_column] = m[i[value_index]+1, out_column] + x[value_index]
        }
    }
    return(m)
}

# ' Create metacells from dgTMatrix sparse matrix
sum_dgTMatrix_dge_by_cluster<-function(idents, raw_dge) {
    m = create_sum_by_cluster_matrix(idents, raw_dge, "dgTMatrix")
    x = raw_dge@x # the non-zero elements of the sparse matrix
    i = raw_dge@i # zero-based row number of corresponding value in x
    j = raw_dge@j # zero-based column number of corresponding value in x
    for (index in 1:length(x)) {
        out_column = as.integer(idents[j[index]+1])
        m[i[index]+1, out_column] = m[i[index]+1, out_column] + x[index]
    }
    return(m)
}


#' Create metacells that have the sum of digital expression of the cells in each class
#' @param idents factor of cell assignments.  length(idents) must == ncol(raw_dge)
#' @param raw_dge digital expression.  May be sparse or dense.  Columns are cells, rows are genes
#' @return matrix with columns => cluster IDs from levels(idents), rows => genes
#' @export
sum_dge_by_cluster=function(idents, raw_dge) {
    if (length(idents) != ncol(raw_dge)) {
        stop(sprintf("length(idents)(%d) != ncol(raw_dge)(%d)", length(idents), ncol(raw_dge)))
    }
    if (is(raw_dge, "dgTMatrix")) {
        return(sum_dgTMatrix_dge_by_cluster(idents, raw_dge))
    } else if (is(raw_dge, "dgCMatrix")) {
        return(sum_dgCMatrix_dge_by_cluster(idents, raw_dge))
    } else {
        # Dense matrix or some other kind of sparse
        SumByCellClass = function(x, cell_class) tapply( x, cell_class, sum )
        ret = apply(raw_dge, 1, SumByCellClass, cell_class=idents)
        # if there is only one cluster, apply returns a vector instead of a matrix, so construct a matrix in that case.
        if (nlevels(idents) == 1) {
            tret = cbind(ret)
            colnames(tret) = levels(idents)
        } else {
            tret = t(ret)
        }
        return(tret)
    }
}

#' Aligns a pair of two similar matrices by rows and columns then runs specified function on aligned matrices
#'
#' Can be combined with Reduce to apply over a list of similarly formed matrices. Very use
#' 
#' @export
#' @param m1 numeric matrix with row and column names
#' @param m2 numeric matrix with row and column names
#' @param FUN function to be applied over conformed matrices
#' @param method.row method by which to merge rows. If 'intersect' (default) returns a matrix with shared rows (by name) between two matrices;
#'                   if 'union' returns a matrix with rows from either m1 or m2, assuming a value of 0 for the matrix it was not present in;
#'                   if 'strict' will validate that m1 and m2 have exactly the same rows otherwise will fail (order doesn't matter though)
#' @param method.col method by which to merge columns. If 'intersect' (default) returns a matrix with shared columns (by name) between two matrices;
#'                   if 'union' returns a matrix with columns from either m1 or m2, assuming a value of 0 for the matrix it was not present in;
#'                   if 'strict' will validate that m1 and m2 have exactly the same columns otherwise will fail (order doesn't matter though)
apply_to_aligned_matrix_pair=function(m1,m2,FUN,method.row=c('intersect','union','strict','ignore'), method.col=c('intersect','union','strict','ignore')) {
    if (!(is.matrix(m1) & is.matrix(m2))) stop("m1 and m2 must both be matrices")
    method.row=match.arg(method.row)
    method.col=match.arg(method.col)
    rn1=rownames(m1)
    rn2=rownames(m2)
    cn1=colnames(m1)
    cn2=colnames(m2)
    
    if (method.row == 'intersect')  {
        # Subset m1 and m2 to only have shared rows
        row.idx=intersect(rn1,rn2)
        m1=m1[row.idx,]
        m2=m2[row.idx,] 
    } 
    else if (method.row == 'union') { 
        # rows not in m2 that will be set to 0
        rx2=setdiff(rn1,rn2) 
        # rows not in m1 that will be set to 0
        rx1=setdiff(rn2,rn1) 
        # defines row order after alignment
        row.idx=union(rn1,rn2) 
        if (length(rx1) > 0) {
            # make length(rx1) new rows of zeros
            x=matrix(0,length(rx1),ncol(m1)) 
            # name these rows by the names in rx1
            rownames(x)=rx1 
            # add rows to m1
            m1=rbind(m1,x) 
        }
        if (length(rx2) > 0) {
            # make length(rx2) new rows of zeros
            x=matrix(0,length(rx2),ncol(m2))
            # name these rows by the names in rx2
            rownames(x)=rx2
            # add rows to m2
            m2=rbind(m2,x)
        }
        # re-order rows of both matrices
        m1=m1[row.idx,]
        m2=m2[row.idx,]
    } else if (method.row == 'strict') {
        # Ensures that rows of m1 and m2 are the same
        if ( !all(sort(rn1) == sort(rn2)) ) {
            stop("Set of rows are not equivalent")
        }  else { 
            # Give two matrices the same order
            # Assumes order of the first matrix
            row.idx=rn1
            m1=m1[row.idx,]
            m2=m2[row.idx,]
        }
    }
    
    if (method.col == 'intersect')  { 
        # Subsets both to shared columns
        col.idx=intersect(cn1,cn2)
        m1=m1[,col.idx]
        m2=m2[,col.idx]
    } else if (method.col == 'union') { 
        # Columns not in m2
        cx2=setdiff(cn1,cn2)
        # Columns not in m1
        cx1=setdiff(cn2,cn1)
        col.idx=union(cn1,cn2)
        if (length(cx1) > 0) {
            # Make length(cx1) new columns
            x=matrix(0,nrow(m1),length(cx1))
            # Name these columns by cx1
            colnames(x)=cx1 # Add columns to m1
            m1=cbind(m1,x)
        }
        if (length(cx2) > 0) {
            # Make length(cx2) new columns
            x=matrix(0,nrow(m2),length(cx2))
            # Name these columns by cx2
            colnames(x)=cx2
            # Add columns to m2
            m2=cbind(m2,x)
        }
        # Ensure two matrices have the same order of columns
        m1=m1[,col.idx]
        m2=m2[,col.idx]
    } else if (method.col == 'strict') {
        # Makes sure both have all the same columns
        if (!all(sort(cn1) == sort(cn2))) {
            stop("Set of columns are not equivalent")
        } else {
            # Ensures order of columns
            col.idx=sort(union(cn1,cn2))
            m1=m1[col.idx,]
            m2=m2[col.idx,]
        }
    }
    
    if (is.function(FUN))
        FUN(m1,m2)
    else if (is.character(FUN))
        do.call(FUN,list(m1,m2))
}

#' Sum metacells or DGEs with possibly different sets of genes/columns
#' 
#' @export
#' @param ... list of matrices as first n arguments
#' @param metacells if not missing will be used as list of metacells to sum
#' @param method.row method by which to merge rows. If 'intersect' (default) returns a matrix with shared rows (by name) between two matrices;
#'                   if 'union' returns a matrix with rows from either m1 or m2, assuming a value of 0 for the matrix it was not present in;
#'                   if 'strict' will validate that m1 and m2 have exactly the same rows otherwise will fail (order doesn't matter though)
#' @param method.col method by which to merge columns. If 'intersect' (default) returns a matrix with shared columns (by name) between two matrices;
#'                   if 'union' returns a matrix with columns from either m1 or m2, assuming a value of 0 for the matrix it was not present in;
#'                   if 'strict' will validate that m1 and m2 have exactly the same columns otherwise will fail (order doesn't matter though)
sum_metacells=function(..., metacells, method.row=c('intersect','union','strict'), method.col=c('intersect','union','strict')) {
    if (missing(metacells))
        metacells=list(...)
    else if (!is.list(metacells))
        stop("metacells argument must be a list")
    Reduce(function(m1,m2) {
        apply_to_aligned_matrix_pair(m1,m2,FUN=`+`,method.row=method.row, method.col=method.col)
    }, metacells)
}

#' Force package to be unloaded.  This can be useful when in debug/build cycle.
#' @export
unload_package<-function(package_name) {
    require(package_name, character.only = TRUE)
    detach(paste0('package:', package_name), unload = TRUE, character.only = TRUE)
}
